from datetime import datetime
from typing import (
    Optional,
)
from pydantic import BaseModel, Field
from enum import Enum


class StatusEnum(str, Enum):
    pending = "pending"
    confirmed = "confirmed"


class DatasetMeta(BaseModel):
    slug: str
    name: str
    description: str
    size: int
    created_at: datetime
    price: Optional[int]
    status: StatusEnum
    license_: Optional[str] = Field(None, alias="license")


class Dataset:
    """A Dataset backed by an Arrow table."""

    def __init__(
        self,
        meta: Optional[DatasetMeta] = None,
    ):
        self._meta = meta



from .api import API as InternalAPI
from .display import Display as InternalDisplay
import os

NOTIA_ENDPOINT = os.environ.get("NOTIA_ENDPOINT", "https://notia.api.notia.ai")
NOTIA_WEB = os.environ.get("NOTIA_WEB", "https://notia.ai")
Api = InternalAPI(api_url=NOTIA_ENDPOINT)
Display = InternalDisplay()

NOTIA_CACHE = "~/.cache/notia/datasets"
NOTIA_CACHE = os.path.expanduser(os.getenv("NOTIA_CACHE", NOTIA_CACHE))

EXTRACTED_DATASETS_DIR = "extracted"
DOWNLOADED_DATASETS_DIR = "downloads"



from notia.models import DatasetMeta
from rich.console import Console
from rich.table import Table
from rich import box
from typing import List, Optional


class Display:
    def __init__(self) -> None:
        self._console = Console()

    def log(self, msg_obj=None) -> None:
        self._console.print(msg_obj, style="bold green")

    def warning(self, msg_obj=None) -> None:
        self._console.print(msg_obj, style="bold yellow")

    def error(self, msg_obj=None) -> None:
        self._console.print(msg_obj, style="bold red")

    def log_styled(self, msg_obj=None, style: Optional[str] = None) -> None:
        self._console.print(msg_obj, style=style)

    def datasetsAsTable(self, datasets: List[DatasetMeta], web_url) -> None:
        table = Table(show_header=True, expand=True, box=box.ROUNDED)
        table.add_column("ID")
        table.add_column("Name")
        table.add_column("Size", justify="right")
        for dataset in datasets:
            table.add_row(
                dataset.slug,
                self._displayItemLink(web_url, dataset.name, dataset.slug),
                self._displayFileSize(float(dataset.size)),
            )

        self._console.print(table)

    def ordersAsTable(self, orders) -> None:
        table = Table(show_header=True, header_style="bold blue")
        table.add_column("Name")
        table.add_column("Size")
        for order in orders:
            table.add_row(order["purchase_name"], order["purchase_size"])
        self._console.print(table)

    def _displayFileSize(self, filesize: float, suffix="B") -> str:
        for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
            if abs(filesize) < 1024.0:
                return f"{filesize:3.1f}{unit}{suffix}"
            filesize /= 1024.0
        return f"{filesize:.1f}Yi{suffix}"

    def _displayItemLink(self, base_url: str, name: str, slug: str) -> str:
        return f"[link={base_url}/dataset/{slug}]{name}[/link]"



from notia import apikey
from .display import Display
from .config import NOTIA_ENDPOINT, NOTIA_WEB


class _NotiaLogin:
    def __init__(self):
        self._api_url = NOTIA_ENDPOINT
        self._web_url = NOTIA_WEB
        self._key = None
        self._relogin = None
        self._display = Display()

    def setup(self, kwargs):
        self._relogin = kwargs.pop("relogin", None)

    def login(self):
        """
        Login checks is API key is configured, if not prompts user to input
        No actual logging in is done.
        """
        apikey_configured = self.is_apikey_configured()
        if self._relogin:
            apikey_configured = False
        if not apikey_configured:
            return False

        return apikey_configured

    def update_session(self, key: str) -> None:
        from .config import Api

        Api.api_key = key

    def is_apikey_configured(self):
        return apikey.api_key_exists(self._api_url) is not None

    def configure_api_key(self, key):
        self._display.warning(
            (
                "If you are specifying your api key in code, ensure this code is"
                "not shared publically."
                "Consider setting the NOTIA_API_KEY environment variable"
            )
        )
        apikey.write_key(self._api_url, key)
        self.update_session(key)
        self._key = key

    def prompt_api_key(self):
        key = apikey.prompt_api_key(self._web_url, self._api_url)
        self.update_session(key)
        self._key = key


def login(key=None, relogin=None):
    """
    Log in to Notia.
    Arguments:
        key: (string, optional) authentication key.
        relogin: (bool, optional) If true, will re-prompt for API key.
    Returns:
        bool: if key is configured
    Raises:
        UsageError - if api_key can not configured and no tty
    """
    kwargs = dict(locals())
    configured = _login(**kwargs)
    return True if configured else False


def _login(
    key=None,
    relogin=None,
):
    kwargs = dict(locals())

    nlogin = _NotiaLogin()

    nlogin.setup(kwargs)

    logged_in = nlogin.login()

    key = kwargs.get("key")
    if key:
        nlogin.configure_api_key(key)

    if logged_in:
        return logged_in

    if not key:
        nlogin.prompt_api_key()

    return nlogin._key or False



from zipfile import BadZipfile
import requests
from typing import Optional
from notia.apikey import api_key_exists
from notia.models.dataset import DatasetMeta
from .display import Display


class API:
    """
    Class representing calls to the Notia.ai API
    """

    def __init__(self, api_url: str) -> None:
        self._session = requests.Session()
        self._api_url = api_url
        self._api_key = api_key_exists(self._api_url)
        self._display = Display()

    @property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, new_key):
        self._api_key = new_key

    def Search(self, term=None, web_url=None) -> None:
        try:
            response = self._RequestUrl(
                self._api_url + "/datasets/filter", "GET", params={"search_query": term}
            )
            response_results = response.json()["results"]
            dataset_list = [
                DatasetMeta(**dataset["dataset"]) for dataset in response_results
            ]
            self._display.datasetsAsTable(dataset_list, web_url)
        except requests.exceptions.RequestException:
            self._display.error(
                (
                    "Failed to fetch search resulsts, please ensure you "
                    "have the correct login token, or contact support@notia.ai"
                )
            )
        except Exception as err:
            self._display.error(f"{err}")

    def Orders(self) -> None:
        try:
            response = self._RequestUrl(f"{self._api_url}/orders", "GET")
            self._display.ordersAsTable(response.json())
        except requests.exceptions.RequestException:
            self._display.error(
                (
                    "Failed to fetch user orders, please ensure you "
                    "have the correct login token, or contact support@notia.ai"
                )
            )
            raise

    def GetPresigned(self, slug: str) -> Optional[str]:
        request_url = self._api_url + "/api/datasets/presign"
        try:
            response = self._RequestUrl(request_url, "GET", params={"slug": slug})
            response.raise_for_status()
            presigned_url = response.json()["url"]
            return presigned_url
        except requests.exceptions.RequestException:
            self._display.error(
                (
                    "Failed to download dataset, please ensure you "
                    "have purchased this ID or contact support@notia.ai"
                )
            )
        except BadZipfile:
            self._display.error(
                (
                    "Error unzipping dataset, please manually inspect "
                    "the downloaded file or contact support@notia.ai"
                )
            )
            raise

    def _RequestUrl(
        self, url: str, verb: str, params: Optional[dict] = None, json=None
    ) -> requests.Response:
        if verb == "POST":
            return self._session.post(url, auth=(self._api_key, ""))
        elif verb == "GET":
            return self._session.get(url, params=params, auth=(self._api_key, ""))
        else:
            raise ValueError("Invalid HTTP operation provided")


def search(term=None) -> None:
    from .config import Api, NOTIA_WEB

    Api.Search(term, NOTIA_WEB)


def my_datasets() -> None:
    from .config import Api

    Api.Orders()



import os
from urllib.parse import urlparse
import textwrap
import stat
import getpass

from .display import Display
from rich.panel import Panel
from rich import box

display = Display()


def _find_netrc_api_key(url, raise_errors=False):
    NETRC_FILES = (".netrc", "_netrc")
    netrc_file = os.environ.get("NETRC")
    if netrc_file is not None:
        netrc_locations = (netrc_file,)
    else:
        netrc_locations = ("~/{}".format(f) for f in NETRC_FILES)

    try:
        from netrc import netrc, NetrcParseError

        netrc_path = None

        for f in netrc_locations:
            try:
                loc = os.path.expanduser(f)
            except KeyError:
                return

            if os.path.exists(loc):
                netrc_path = loc
                break

        if netrc_path is None:
            return

        ri = urlparse(url)

        host = ri.netloc.split(":")[0]

        try:
            _netrc = netrc(netrc_path).authenticators(host)
            if _netrc:
                login_i = 0 if _netrc[0] else 1
                return (_netrc[login_i], _netrc[2])
        except (NetrcParseError, IOError):
            if raise_errors:
                raise

    except (ImportError, AttributeError):
        pass


def prompt_api_key(web_url: str, api_url: str):
    api_prompt = Panel(
        (
            f"You can find your API key :key: in your browser"
            f" [blue underline bold][link={web_url}/authorize]here[/link][/blue underline bold] "
        ),
        box=box.ROUNDED,
    )
    display.log_styled(api_prompt, style="yellow")
    display.log(
        "Paste the API key from your profile and hit enter, or press ctrl+c to quit: "
    )
    api_key = getpass.getpass()
    write_netrc(api_url, "user", api_key)
    return api_key


def api_key_exists(api_url: str):
    auth = _find_netrc_api_key(api_url, True)
    key = None
    if auth:
        key = auth[-1]
    # Environment should take precedence
    if os.getenv("NOTIA_API_KEY"):
        key = os.environ["NOTIA_API_KEY"]
    return key


def write_key(api_url, key):
    if not key:
        raise ValueError("No API key specified.")

    _, suffix = key.split("-", 1) if "-" in key else ("", key)

    if len(suffix) != 40:
        raise ValueError("API key must be 40 characters long, yours was %s" % len(key))

    write_netrc(api_url, "user", key)


def write_netrc(host: str, entity: str, key: str):
    """Add our host and key to .netrc"""
    if len(key) != 40:
        display.error(
            f"API-key must be exactly 40 characters long: {key} ({len(key)} chars)"
        )
        return None
    try:
        normalized_host = urlparse(host).netloc.split(":")[0]
        if normalized_host != "localhost" and "." not in normalized_host:
            display.error(
                f"Host must be a url in the form https://some.address.com, received {host}"
            )
            return None
        display.log(
            (
                f"Appending key for {normalized_host} "
                f"to your netrc file: {os.path.expanduser('~/.netrc')}"
            )
        )
        machine_line = "machine %s" % normalized_host
        path = os.path.expanduser("~/.netrc")
        orig_lines = None
        try:
            with open(path) as f:
                orig_lines = f.read().strip().split("\n")
        except IOError:
            pass
        with open(path, "w") as f:
            if orig_lines:
                # delete this machine from the file if it's already there.
                skip = 0
                for line in orig_lines:
                    # we fix invalid netrc files with an empty host that we wrote before
                    # verifying host...
                    if line == "machine " or machine_line in line:
                        skip = 2
                    elif skip:
                        skip -= 1
                    else:
                        f.write("%s\n" % line)
            f.write(
                textwrap.dedent(
                    """\
            machine {host}
              login {entity}
              password {key}
            """
                ).format(host=normalized_host, entity=entity, key=key)
            )
        os.chmod(os.path.expanduser("~/.netrc"), stat.S_IRUSR | stat.S_IWUSR)
        return True
    except IOError:
        display.error("Unable to read ~/.netrc")
        return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['notia', 'notia.errors', 'notia.models']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.0,<4.0.0',
 'click>=8.0.3,<9.0.0',
 'ipywidgets>=7.7.0,<8.0.0',
 'pandas>=1.3,<2.0',
 'pydantic>=1.9.0,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'rich>=12.2.0,<13.0.0']

setup_kwargs = {
    'name': 'notia',
    'version': '0.5.0',
    'description': 'A library for interacting with the Notia.ai platform',
    'long_description': '<div align="center">\n<img src="https://raw.githubusercontent.com/notia-ai/client/master/resources/imgs/notia-gh-logo.png" width=60%/><br/>  \n</div>\n<p align="center">\n<a href="https://pypi.org/project/notia/">\n    <img alt="PyPi Release" src="https://img.shields.io/pypi/v/notia">\n</a> \n</p>\n\n\n---\n\nUse Notia to supercharge your models with the latest training data.\n\n-   Browse 100s of datasets from top companies and institutions.\n-   Directly integrates with Jupyter for easy sharing and distribution.\n-   Focus on the data science. Spend less time manually scraping and cleaning data.\n-   Reproduce any experiment without any Google Drive, Dropbox etc.\n\n## ⚡️ Quick Install\n\nTo install notia, simply:\n\n```bash\npip install notia\n```\n\n## Usage\n\n```python\nimport notia\nnotia.login()\n\nnotia.search("Yelp")\n\ntrain_df = notia.load_dataset("XXXXX", split="train")\n```\n\n## Contributing\n\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n',
    'author': 'Notia.ai',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://notia.ai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



from typing import Optional
import requests

from .display import Display
import time
from pathlib import Path
import os
import copy
import pandas as pd
from rich.progress import track
import tempfile
from functools import partial
from zipfile import BadZipfile, ZipFile
import shutil
import json
from .config import NOTIA_CACHE, Api, EXTRACTED_DATASETS_DIR, DOWNLOADED_DATASETS_DIR


class DownloadManager:
    """A class for managing caching & downloads from Notia.ai"""

    def __init__(self) -> None:
        self._display = Display()
        # cache_dir ~/.cache/notia/datasets
        self.cache_dir = NOTIA_CACHE
        # downloads_dir ~/.cache/notia/datasets/downloads
        self.downloads_dir = os.path.join(self.cache_dir, DOWNLOADED_DATASETS_DIR)
        # extract_dir ~/.cache/notia/datasets/extracted
        self.extract_dir = os.path.join(self.cache_dir, EXTRACTED_DATASETS_DIR)

    def get_from_cache(self, slug: str, force_download=False):
        if self.cache_dir is None:
            self.cache_dir = NOTIA_CACHE

        if isinstance(self.cache_dir, Path):
            self.cache_dir = str(self.cache_dir)

        os.makedirs(self.cache_dir, exist_ok=True)

        cache_path = os.path.join(self.cache_dir, slug)

        if os.path.exists(cache_path) and not force_download:
            return cache_path

        os.makedirs(self.cache_dir, exist_ok=True)
        os.makedirs(self.downloads_dir, exist_ok=True)

        temp_file_manager = partial(
            tempfile.NamedTemporaryFile, dir=self.downloads_dir, delete=False
        )

        try:
            with temp_file_manager() as temp_file:
                self.download_from_s3(temp_file, slug)
                shutil.move(temp_file.name, os.path.join(self.downloads_dir, slug))
                temp_file.close()
        except requests.exceptions.RequestException:
            # propagate upwards
            raise

        # extract the file from the download dir to the extract dir
        self.extract(
            os.path.join(self.downloads_dir, slug), os.path.join(self.extract_dir, slug)
        )

        shutil.move(os.path.join(self.extract_dir, slug), cache_path)
        os.remove(os.path.join(self.downloads_dir, slug))
        return cache_path

    def extract(self, input_path: str, output_path: str):
        try:
            os.makedirs(output_path, exist_ok=True)
            with ZipFile(input_path, "r") as zip_file:
                zip_file.extractall(path=output_path)
                zip_file.close()
        except BadZipfile:
            # propagate upwards
            raise

    def download_from_s3(self, temp_file, slug: str) -> None:
        presigned_url = Api.GetPresigned(slug)
        if presigned_url is not None:
            self._display.log(
                (
                    f"{slug} not found in cache or force_download set to True,"
                    f"downloading to {temp_file.name}"
                )
            )
            self.http_get(presigned_url, temp_file)
        else:
            raise ValueError(f"Unable to fetch resource at {slug}")

    def http_get(
        self, url, temp_file, resume_size=0, headers=None, timeout=100.0, max_retries=0
    ):
        headers = copy.deepcopy(headers) or {}
        if resume_size > 0:
            headers["Range"] = f"bytes={resume_size:d}-"
        response = self.request_with_retry(
            method="GET",
            url=url,
            max_retries=max_retries,
            timeout=timeout,
        )
        if response.status_code == 416:  # Range not satisfiable
            return

        content_length = response.headers.get("Content-Length")
        total = (
            resume_size + int(content_length) if content_length is not None else None
        )
        for chunk in track(
            response.iter_content(chunk_size=1024),
            description="Downloading...",
            total=total,
        ):
            if chunk:  # filter out keep-alive new chunks
                temp_file.write(chunk)

    def request_with_retry(
        self,
        method: str,
        url: str,
        max_retries: int = 0,
        base_wait_time: float = 0.5,
        max_wait_time: float = 2,
        timeout: float = 10.0,
        **params,
    ) -> requests.Response:
        """Wrapper around requests to retry in case it fails with a
           ConnectTimeout, with exponential backoff.

        Args:
            method (str): HTTP method, such as 'GET' or 'HEAD'.
            url (str): The URL of the resource to fetch.
            max_retries (int): Maximum number of retries, defaults to
                               0 (no retries).
            base_wait_time (float): Duration (in seconds) to wait before
                                    retrying the first time. Wait time between
                                    retries then grows exponentially,
                                    capped by max_wait_time.
            max_wait_time (float): Maximum amount of time between
                                   two retries, in seconds.
            **params: Params to pass to :obj:`requests.request`.
        """
        tries, success = 0, False
        while not success:
            tries += 1
            try:
                response = requests.request(
                    method=method.upper(), url=url, timeout=timeout, **params
                )
                success = True
            except (
                requests.exceptions.ConnectTimeout,
                requests.exceptions.ConnectionError,
            ) as err:
                if tries > max_retries:
                    raise err
                else:
                    self._display.log(
                        (
                            f"{method} request to {url} timed out,"
                            "retrying... [{tries/max_retries}]"
                        )
                    )
                    sleep_time = min(
                        max_wait_time, base_wait_time * 2 ** (tries - 1)
                    )  # Exponential backoff
                    time.sleep(sleep_time)
        return response


def _handle_tsv(tsv_path: str):
    return pd.read_csv(tsv_path, sep="\t")


def _handle_csv(csv_path: str):
    return pd.read_csv(csv_path)


def _handle_json(json_path: str):
    return json.loads(json_path)


def _handle_split(local_path: str, ext: str, split: str):
    fpath = os.path.join(local_path, split)
    if os.path.exists(f"{fpath}.csv"):
        return _handle_csv(f"{fpath}.csv")
    elif os.path.exists(f"{fpath}.tsv"):
        return _handle_tsv(f"{fpath}.tsv")
    elif os.path.exists(f"{fpath}.json"):
        return _handle_json(f"{fpath}.json")
    else:
        raise ValueError(
            (
                f"Load was true but {ext} was provided. Notia can automatically "
                "load CSV, JSON or TSV files. Please set load=False and manually "
                "load the file."
            )
        )


def load_dataset(ID: str, split: Optional[str] = None, load: Optional[bool] = True):
    download_manager = DownloadManager()
    local_path = download_manager.get_from_cache(ID)
    if not load:
        return local_path

    if split:
        # this should know the ext
        return _handle_split(local_path, "", split)
    else:
        datasets = []
        for file in os.listdir(local_path):
            # we pass the local path and the file stem so we can match on ext
            datasets.append(
                _handle_split(local_path, Path(file).suffix, Path(file).stem)
            )
        if len(datasets) == 1:
            # no need to tuple if single elem
            return datasets[0]
        else:
            return tuple(datasets)
